@@ -22,6 +22,8 @@
 #include <string.h>
 #include <libgen.h>
 
+#include <arpa/inet.h>
+
 #include <netinet/in.h>
 #include <netinet/in_systm.h>
 #include <netinet/ip.h>
@@ -31,6 +33,10 @@
 #include <netinet/icmp6.h>
 #endif
 
+#ifdef HAVE_LINUX_FILTER_H
+#include <linux/filter.h>
+#endif
+
 #ifdef __CYGWIN__
 #include "../include/cygwin.h"
 #endif
@@ -43,11 +49,11 @@
 #endif /* HAVE_SENDMMSG || HAVE_RECVMMSG */
 #endif /* ENABLE_MMSG */
 
-static const size_t   MAX_IPV4_HDR_SIZE   = 60;
-static const uint32_t CALIBRATION_CYCLES  = 100,
-                      PKT_BURST_PRECISION = 1000;
-static const int64_t  MAX_PKT_TIME_SEC    = 8589934592,
-                      MAX_PKT_TIME_NSEC   = 999999999;
+static const size_t   MAX_IPV4_HDR_SIZE  = 60;
+static const uint32_t CALIBRATION_CYCLES = 100,
+                      PKT_BURST_SCALE    = 1000;
+static const int64_t  MAX_PKT_TIME_SEC   = 8589934592,
+                      MAX_PKT_TIME_NSEC  = 999999999;
 
 static char *prog_name;
 
@@ -90,7 +96,7 @@ static int64_t ts_sub(const struct timespec *ts1, const struct timespec *ts2)
     return ((int64_t)ts1->tv_sec - (int64_t)ts2->tv_sec) * 1000000 + (ts1->tv_nsec - ts2->tv_nsec) / 1000;
 }
 
-static uint16_t cksum(const char *packet, size_t size)
+static uint16_t cksum(const char *data, size_t size)
 {
     uint32_t sum = 0;
 
@@ -98,9 +104,9 @@ static uint16_t cksum(const char *packet, size_t size)
         uint16_t u16 = 0;
 
         if (i < size - 1) {
-            memcpy(&u16, &packet[i], 2);
+            memcpy(&u16, &data[i], 2);
         } else {
-            memcpy(&u16, &packet[i], 1);
+            memcpy(&u16, &data[i], 1);
         }
 
         sum += u16;
@@ -114,32 +120,41 @@ static uint16_t cksum(const char *packet, size_t size)
 
 static int64_t calibrate_timer(void)
 {
-    int64_t sum = 0;
+    uint32_t successful_cycles = 0;
+    int64_t  sum               = 0;
 
     for (uint32_t i = 0; i < CALIBRATION_CYCLES; i++) {
-        struct timespec begin, end;
+        struct timespec before;
 
-        int n = -1;
+        get_time(&before);
 
-        while (n < 0) {
-            get_time(&begin);
+        struct timeval timeout = {.tv_sec = 0, .tv_usec = 10};
 
-            struct timeval timeout = {.tv_sec = 0, .tv_usec = 10};
+        if (select(0, NULL, NULL, NULL, &timeout) < 0) {
+            fprintf(stderr, "%s: select() failed: %s\n", prog_name, strerror(errno));
+        } else {
+            struct timespec after;
 
-            n = select(0, NULL, NULL, NULL, &timeout);
+            get_time(&after);
+
+            successful_cycles += 1;
+            sum               += ts_sub(&after, &before);
         }
-
-        get_time(&end);
-
-        sum += ts_sub(&end, &begin);
     }
 
-    return sum / CALIBRATION_CYCLES;
+    return successful_cycles > 0 ? sum / successful_cycles : 0;
 }
 
-static void prepare_ping4(char *packet, size_t pkt_size, uint16_t ident, bool insert_timestamp, uint64_t *transmitted_count, uint64_t *transmitted_volume)
+static void prepare_ping4(char *packet, size_t pkt_size, uint16_t ident, bool insert_timestamp,
+                          uint64_t *transmitted_count, uint64_t *transmitted_volume)
 {
-    struct icmp icmp4 = {.icmp_type = ICMP_ECHO, .icmp_code = 0, .icmp_cksum = 0, .icmp_id = ident, .icmp_seq = htons(*transmitted_count)};
+    struct icmp icmp4 = {
+        .icmp_type  = ICMP_ECHO,
+        .icmp_code  = 0,
+        .icmp_cksum = 0,
+        .icmp_id    = htons(ident),
+        .icmp_seq   = htons(*transmitted_count)
+    };
 
     memcpy(packet, &icmp4, sizeof(icmp4));
 
@@ -159,13 +174,20 @@ static void prepare_ping4(char *packet, size_t pkt_size, uint16_t ident, bool in
 
     memcpy(&packet[offsetof(struct icmp, icmp_cksum)], &icmp4.icmp_cksum, sizeof(icmp4.icmp_cksum));
 
-    (*transmitted_count)++;
-    (*transmitted_volume) += pkt_size;
+    *transmitted_count  += 1;
+    *transmitted_volume += pkt_size;
 }
 
-static void prepare_ping6(char *packet, size_t pkt_size, uint16_t ident, bool insert_timestamp, uint64_t *transmitted_count, uint64_t *transmitted_volume)
+static void prepare_ping6(char *packet, size_t pkt_size, uint16_t ident, bool insert_timestamp,
+                          uint64_t *transmitted_count, uint64_t *transmitted_volume)
 {
-    struct icmp6_hdr icmp6 = {.icmp6_type = ICMP6_ECHO_REQUEST, .icmp6_code = 0, .icmp6_cksum = 0, .icmp6_id = ident, .icmp6_seq = htons(*transmitted_count)};
+    struct icmp6_hdr icmp6 = {
+        .icmp6_type  = ICMP6_ECHO_REQUEST,
+        .icmp6_code  = 0,
+        .icmp6_cksum = 0,
+        .icmp6_id    = htons(ident),
+        .icmp6_seq   = htons(*transmitted_count)
+    };
 
     memcpy(packet, &icmp6, sizeof(icmp6));
 
@@ -177,13 +199,14 @@ static void prepare_ping6(char *packet, size_t pkt_size, uint16_t ident, bool in
         memcpy(&packet[sizeof(icmp6)], &pkt_time, sizeof(pkt_time));
     }
 
-    (*transmitted_count)++;
-    (*transmitted_volume) += pkt_size;
+    *transmitted_count  += 1;
+    *transmitted_volume += pkt_size;
 }
 
 #if defined(ENABLE_MMSG) && defined(HAVE_SENDMMSG)
 
-static void sendmmsg_ping(bool ipv4_mode, int sock, const struct addrinfo *to_ai, size_t pkt_size, uint16_t ident, uint64_t pkt_count, uint64_t *transmitted_count, uint64_t *transmitted_volume)
+static void sendmmsg_ping(bool ipv4_mode, int sock, const struct addrinfo *to_ai, size_t pkt_size, uint16_t ident,
+                          uint64_t pkt_count, uint64_t *transmitted_count, uint64_t *transmitted_volume)
 {
     for (uint64_t i = 0; i < pkt_count; i = i + MAX_MMSG_VLEN) {
         static char packets[MAX_MMSG_VLEN][IP_MAXPACKET];
@@ -227,16 +250,17 @@ static void sendmmsg_ping(bool ipv4_mode, int sock, const struct addrinfo *to_ai
 
 #else /* ENABLE_MMSG && HAVE_SENDMMSG */
 
-static void send_ping(bool ipv4_mode, int sock, const struct addrinfo *to_ai, size_t pkt_size, uint16_t ident, bool first_in_burst, uint64_t *transmitted_count, uint64_t *transmitted_volume)
+static void send_ping(bool ipv4_mode, int sock, const struct addrinfo *to_ai, size_t pkt_size, uint16_t ident,
+                      bool insert_timestamp, uint64_t *transmitted_count, uint64_t *transmitted_volume)
 {
     static char packet[IP_MAXPACKET];
 
     memset(packet, 0, pkt_size);
 
     if (ipv4_mode) {
-        prepare_ping4(packet, pkt_size, ident, first_in_burst, transmitted_count, transmitted_volume);
+        prepare_ping4(packet, pkt_size, ident, insert_timestamp, transmitted_count, transmitted_volume);
     } else {
-        prepare_ping6(packet, pkt_size, ident, first_in_burst, transmitted_count, transmitted_volume);
+        prepare_ping6(packet, pkt_size, ident, insert_timestamp, transmitted_count, transmitted_volume);
     }
 
     ssize_t res = sendto(sock, packet, pkt_size, 0, to_ai->ai_addr, to_ai->ai_addrlen);
@@ -250,53 +274,56 @@ static void send_ping(bool ipv4_mode, int sock, const struct addrinfo *to_ai, si
 
 #endif /* ENABLE_MMSG && HAVE_SENDMMSG */
 
-static void process_ping4(const char *packet, ssize_t pkt_size, uint16_t ident, uint64_t *received_count, uint64_t *received_volume, uint64_t *min_rtt, uint64_t *max_rtt, uint64_t *average_rtt)
+static void process_ping4(const char *packet, ssize_t pkt_size, uint16_t ident, uint64_t *received_count,
+                          uint64_t *received_volume, uint64_t *min_rtt, uint64_t *max_rtt, uint64_t *average_rtt)
 {
     struct ip ip4;
 
     if (pkt_size >= (ssize_t)sizeof(ip4)) {
         memcpy(&ip4, packet, sizeof(ip4));
 
-        size_t hdr_len = ip4.ip_hl << 2;
+        if (ip4.ip_p == IPPROTO_ICMP && (ntohs(ip4.ip_off) & 0x1FFF) == 0) {
+            size_t hdr_len = ip4.ip_hl << 2;
 
-        struct icmp icmp4;
+            struct icmp icmp4;
 
-        if (pkt_size >= (ssize_t)(hdr_len + sizeof(icmp4))) {
-            memcpy(&icmp4, &packet[hdr_len], sizeof(icmp4));
+            if (pkt_size >= (ssize_t)(hdr_len + sizeof(icmp4))) {
+                memcpy(&icmp4, &packet[hdr_len], sizeof(icmp4));
 
-            if (icmp4.icmp_type == ICMP_ECHOREPLY &&
-                icmp4.icmp_id   == ident) {
-                (*received_count)++;
-                (*received_volume) += pkt_size - hdr_len;
+                if (icmp4.icmp_type == ICMP_ECHOREPLY && ntohs(icmp4.icmp_id) == ident) {
+                    *received_count  += 1;
+                    *received_volume += pkt_size - hdr_len;
 
-                struct timespec pkt_time;
+                    struct timespec pkt_time;
 
-                if (pkt_size >= (ssize_t)(hdr_len + sizeof(icmp4) + sizeof(pkt_time))) {
-                    memcpy(&pkt_time, &packet[hdr_len + sizeof(icmp4)], sizeof(pkt_time));
+                    if (pkt_size >= (ssize_t)(hdr_len + sizeof(icmp4) + sizeof(pkt_time))) {
+                        memcpy(&pkt_time, &packet[hdr_len + sizeof(icmp4)], sizeof(pkt_time));
 
-                    if (pkt_time.tv_sec != 0 || pkt_time.tv_nsec != 0) {
-                        if (pkt_time.tv_sec  >= 0 && (int64_t)pkt_time.tv_sec + 0 <= MAX_PKT_TIME_SEC &&
-                            pkt_time.tv_nsec >= 0 &&          pkt_time.tv_nsec    <= MAX_PKT_TIME_NSEC) {
-                            struct timespec now;
+                        if (pkt_time.tv_sec != 0 || pkt_time.tv_nsec != 0) {
+                            if ((int64_t)pkt_time.tv_sec  >= 0 && (int64_t)pkt_time.tv_sec  <= MAX_PKT_TIME_SEC &&
+                                         pkt_time.tv_nsec >= 0 &&          pkt_time.tv_nsec <= MAX_PKT_TIME_NSEC) {
+                                struct timespec now;
 
-                            get_time(&now);
+                                get_time(&now);
 
-                            int64_t rtt = ts_sub(&now, &pkt_time) / 1000;
+                                int64_t rtt = ts_sub(&now, &pkt_time) / 1000;
 
-                            if (rtt >= 0) {
-                                if (*min_rtt > (uint64_t)rtt) {
-                                    *min_rtt = rtt;
+                                if (rtt >= 0) {
+                                    if (*min_rtt > (uint64_t)rtt) {
+                                        *min_rtt = rtt;
+                                    }
+                                    if (*max_rtt < (uint64_t)rtt) {
+                                        *max_rtt = rtt;
+                                    }
+                                    if (*received_count > 0) {
+                                        *average_rtt = (*average_rtt * (*received_count - 1) + rtt) / *received_count;
+                                    }
+                                } else {
+                                    fprintf(stderr, "%s: packet has an invalid timestamp\n", prog_name);
                                 }
-                                if (*max_rtt < (uint64_t)rtt) {
-                                    *max_rtt = rtt;
-                                }
-
-                                *average_rtt = *received_count ? ((*average_rtt * (*received_count - 1)) + rtt) / *received_count : *average_rtt;
                             } else {
                                 fprintf(stderr, "%s: packet has an invalid timestamp\n", prog_name);
                             }
-                        } else {
-                            fprintf(stderr, "%s: packet has an invalid timestamp\n", prog_name);
                         }
                     }
                 }
@@ -305,17 +332,17 @@ static void process_ping4(const char *packet, ssize_t pkt_size, uint16_t ident,
     }
 }
 
-static void process_ping6(const char *packet, ssize_t pkt_size, uint16_t ident, uint64_t *received_count, uint64_t *received_volume, uint64_t *min_rtt, uint64_t *max_rtt, uint64_t *average_rtt)
+static void process_ping6(const char *packet, ssize_t pkt_size, uint16_t ident, uint64_t *received_count,
+                          uint64_t *received_volume, uint64_t *min_rtt, uint64_t *max_rtt, uint64_t *average_rtt)
 {
     struct icmp6_hdr icmp6;
 
     if (pkt_size >= (ssize_t)sizeof(icmp6)) {
         memcpy(&icmp6, packet, sizeof(icmp6));
 
-        if (icmp6.icmp6_type == ICMP6_ECHO_REPLY &&
-            icmp6.icmp6_id   == ident) {
-            (*received_count)++;
-            (*received_volume) += pkt_size;
+        if (icmp6.icmp6_type == ICMP6_ECHO_REPLY && ntohs(icmp6.icmp6_id) == ident) {
+            *received_count  += 1;
+            *received_volume += pkt_size;
 
             struct timespec pkt_time;
 
@@ -323,8 +350,8 @@ static void process_ping6(const char *packet, ssize_t pkt_size, uint16_t ident,
                 memcpy(&pkt_time, &packet[sizeof(icmp6)], sizeof(pkt_time));
 
                 if (pkt_time.tv_sec != 0 || pkt_time.tv_nsec != 0) {
-                    if (pkt_time.tv_sec  >= 0 && (int64_t)pkt_time.tv_sec + 0 <= MAX_PKT_TIME_SEC &&
-                        pkt_time.tv_nsec >= 0 &&          pkt_time.tv_nsec    <= MAX_PKT_TIME_NSEC) {
+                    if ((int64_t)pkt_time.tv_sec  >= 0 && (int64_t)pkt_time.tv_sec  <= MAX_PKT_TIME_SEC &&
+                                 pkt_time.tv_nsec >= 0 &&          pkt_time.tv_nsec <= MAX_PKT_TIME_NSEC) {
                         struct timespec now;
 
                         get_time(&now);
@@ -338,8 +365,9 @@ static void process_ping6(const char *packet, ssize_t pkt_size, uint16_t ident,
                             if (*max_rtt < (uint64_t)rtt) {
                                 *max_rtt = rtt;
                             }
-
-                            *average_rtt = *received_count ? ((*average_rtt * (*received_count - 1)) + rtt) / *received_count : *average_rtt;
+                            if (*received_count > 0) {
+                                *average_rtt = (*average_rtt * (*received_count - 1) + rtt) / *received_count;
+                            }
                         } else {
                             fprintf(stderr, "%s: packet has an invalid timestamp\n", prog_name);
                         }
@@ -354,7 +382,8 @@ static void process_ping6(const char *packet, ssize_t pkt_size, uint16_t ident,
 
 #if defined(ENABLE_MMSG) && defined(HAVE_RECVMMSG)
 
-static bool recvmmsg_ping(bool ipv4_mode, int sock, uint16_t ident, uint64_t *received_count, uint64_t *received_volume, uint64_t *min_rtt, uint64_t *max_rtt, uint64_t *average_rtt)
+static bool recvmmsg_ping(bool ipv4_mode, int sock, uint16_t ident, uint64_t *received_count, uint64_t *received_volume,
+                          uint64_t *min_rtt, uint64_t *max_rtt, uint64_t *average_rtt)
 {
     static char packets[MAX_MMSG_VLEN][IP_MAXPACKET];
 
@@ -394,7 +423,8 @@ static bool recvmmsg_ping(bool ipv4_mode, int sock, uint16_t ident, uint64_t *re
 
 #else /* ENABLE_MMSG && HAVE_RECVMMSG */
 
-static bool recv_ping(bool ipv4_mode, int sock, uint16_t ident, uint64_t *received_count, uint64_t *received_volume, uint64_t *min_rtt, uint64_t *max_rtt, uint64_t *average_rtt)
+static bool recv_ping(bool ipv4_mode, int sock, uint16_t ident, uint64_t *received_count, uint64_t *received_volume,
+                      uint64_t *min_rtt, uint64_t *max_rtt, uint64_t *average_rtt)
 {
     static char packet[IP_MAXPACKET];
 
@@ -498,7 +528,7 @@ int main(int argc, char *argv[])
             case 'r':
                 reporting_period = strtol(optarg, &ep, 0);
 
-                if (*ep || ep == optarg || reporting_period < 0) {
+                if (*ep || ep == optarg) {
                     exit_val = EX_USAGE;
                 }
 
@@ -541,7 +571,8 @@ int main(int argc, char *argv[])
     }
 
     if (exit_val != EX_OK) {
-        fprintf(stderr, "Usage: %s [-4 | -6] [-u buf_size] [-r reporting_period] [-T tos(v4) | traf_class(v6)] [-B bind_addr] -b kbps -s pkt_size -v volume target\n", prog_name);
+        fprintf(stderr, "Usage: %s [-4 | -6] [-u buf_size] [-r reporting_period] [-T tos(v4) | traf_class(v6)] [-B bind_addr]"
+                        " -b kbps -s pkt_size -v volume target\n", prog_name);
 
         exit(exit_val);
     }
@@ -551,6 +582,7 @@ int main(int argc, char *argv[])
             fprintf(stderr, "%s: invalid packet size, should be between %zu and %zu\n", prog_name,
                                                                                         sizeof(struct icmp) + sizeof(struct timespec),
                                                                                         (size_t)IP_MAXPACKET - MAX_IPV4_HDR_SIZE);
+
             exit(EX_USAGE);
         }
     } else {
@@ -558,10 +590,17 @@ int main(int argc, char *argv[])
             fprintf(stderr, "%s: invalid packet size, should be between %zu and %zu\n", prog_name,
                                                                                         sizeof(struct icmp6_hdr) + sizeof(struct timespec),
                                                                                         (size_t)IP_MAXPACKET);
+
             exit(EX_USAGE);
         }
     }
 
+    if (reporting_period < 0) {
+        fprintf(stderr, "%s: invalid reporting period, should be non-negative\n", prog_name);
+
+        exit(EX_USAGE);
+    }
+
     int sock;
 
     if (ipv4_mode) {
@@ -599,7 +638,7 @@ int main(int argc, char *argv[])
 
                 freeaddrinfo(bind_ai);
             } else {
-                exit_val = EX_SOFTWARE;
+                exit_val = EX_NOHOST;
             }
         }
 
@@ -607,8 +646,6 @@ int main(int argc, char *argv[])
             struct addrinfo *to_ai;
 
             if (resolve_name(ipv4_mode, target, &to_ai)) {
-                uint16_t ident = getpid() & 0xFFFF;
-
                 char addr_buf[ipv4_mode ? INET_ADDRSTRLEN : INET6_ADDRSTRLEN];
 
                 if (getnameinfo(to_ai->ai_addr, to_ai->ai_addrlen, addr_buf, sizeof(addr_buf), NULL, 0, NI_NUMERICHOST) != 0) {
@@ -619,31 +656,7 @@ int main(int argc, char *argv[])
                 printf("Target: %s (%s), transfer speed: %" PRIu32 " kbps, packet size: %zu bytes, traffic volume: %" PRIu64 " bytes\n",
                        target, addr_buf, kbps, pkt_size, volume);
 
-                bool     finish             = false;
-                uint64_t min_rtt            = UINT64_MAX,
-                         max_rtt            = 0,
-                         average_rtt        = 0,
-                         transmitted_count  = 0,
-                         received_count     = 0,
-                         transmitted_volume = 0,
-                         received_volume    = 0;
-
-                int64_t interval     = (int64_t)pkt_size * 8000 / kbps,
-                        min_interval = calibrate_timer() * 2; /* Leave space for interval_error adjustments */
-
-                uint64_t pkt_burst;
-
-                if (interval >= min_interval) {
-                    pkt_burst = PKT_BURST_PRECISION * 1;
-                } else if (interval == 0) {
-                    pkt_burst = PKT_BURST_PRECISION * min_interval * kbps / 8000 / pkt_size;
-                    interval  = min_interval;
-                } else {
-                    pkt_burst = PKT_BURST_PRECISION * min_interval / interval;
-                    interval  = min_interval;
-                }
-
-                if (buf_size != 0) {
+                if (buf_size > 0) {
                     if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &buf_size, sizeof(buf_size)) < 0) {
                         fprintf(stderr, "%s: setsockopt(SO_RCVBUF, %u) failed: %s\n", prog_name, buf_size, strerror(errno));
                     }
@@ -662,37 +675,100 @@ int main(int argc, char *argv[])
                     }
                 }
 
-                struct timespec begin, end, report;
+                uint16_t ident = getpid() & 0xFFFF;
 
-                get_time(&begin);
+#if defined(ENABLE_BPF) && defined(HAVE_LINUX_FILTER_H) && defined(SO_ATTACH_FILTER)
+                struct sock_filter filter4[] = {
+                    /* (00) */ {0x30, 0, 0, 0x00000009},     /* ldb  [9]                         - IP Protocol */
+                    /* (01) */ {0x15, 0, 8, IPPROTO_ICMP},   /* jeq  $IPPROTO_ICMP   jt 2  jf 10 - IP Protocol is ICMP */
+                    /* (02) */ {0x28, 0, 0, 0x00000006},     /* ldh  [6]                         - IP Fragment Offset */
+                    /* (03) */ {0x45, 6, 0, 0x00001FFF},     /* jset #0x1FFF         jt 10 jf 4  - IP Fragment Offset is zero */
+                    /* (04) */ {0xB1, 0, 0, 0x00000000},     /* ldxb 4*([0]&0xF)                 - Load IHL*4 to X */
+                    /* (05) */ {0x50, 0, 0, 0x00000000},     /* ldb  [x]                         - ICMP Type */
+                    /* (06) */ {0x15, 0, 3, ICMP_ECHOREPLY}, /* jeq  $ICMP_ECHOREPLY jt 7  jf 10 - ICMP Type is Echo Reply */
+                    /* (07) */ {0x48, 0, 0, 0x00000004},     /* ldh  [x + 4]                     - ICMP Id */
+                    /* (08) */ {0x15, 0, 1, ident},          /* jeq  $ident          jt 9  jf 10 - ICMP Id is equal to ident */
+                    /* (09) */ {0x06, 0, 0, 0x00040000},     /* ret  #0x40000                    - Accept packet */
+                    /* (10) */ {0x06, 0, 0, 0x00000000}      /* ret  #0x0                        - Discard packet */
+                };
+                struct sock_filter filter6[] = {
+                    /* (00) */ {0x30, 0, 0, 0x00000000},       /* ldb [0]                         - ICMPv6 Type */
+                    /* (01) */ {0x15, 0, 3, ICMP6_ECHO_REPLY}, /* jeq $ICMP6_ECHO_REPLY jt 2 jf 5 - ICMPv6 Type is Echo Reply */
+                    /* (02) */ {0x28, 0, 0, 0x00000004},       /* ldh [4]                         - ICMPv6 Id */
+                    /* (03) */ {0x15, 0, 1, ident},            /* jeq $ident            jt 4 jf 5 - ICMPv6 Id is equal to ident */
+                    /* (04) */ {0x06, 0, 0, 0x00040000},       /* ret #0x40000                    - Accept packet */
+                    /* (05) */ {0x06, 0, 0, 0x00000000}        /* ret #0x0                        - Discard packet */
+                };
+
+                struct sock_fprog bpf;
+
+                if (ipv4_mode) {
+                    bpf.len    = sizeof(filter4) / sizeof(filter4[0]);
+                    bpf.filter = filter4;
+                } else {
+                    bpf.len    = sizeof(filter6) / sizeof(filter6[0]);
+                    bpf.filter = filter6;
+                }
+
+                if (setsockopt(sock, SOL_SOCKET, SO_ATTACH_FILTER, &bpf, sizeof(bpf)) < 0) {
+                    fprintf(stderr, "%s: setsockopt(SO_ATTACH_FILTER) failed: %s\n", prog_name, strerror(errno));
+                }
+#endif /* ENABLE_BPF && HAVE_LINUX_FILTER_H && SO_ATTACH_FILTER */
+
+                int64_t interval     = (int64_t)pkt_size * 8000 / kbps,
+                        min_interval = calibrate_timer() * 2; /* Leave space for interval_error adjustments */
+
+                uint64_t pkt_burst;
+
+                if (interval >= min_interval) {
+                    pkt_burst = PKT_BURST_SCALE * 1;
+                } else if (interval == 0) {
+                    pkt_burst = PKT_BURST_SCALE * min_interval * kbps / 8000 / pkt_size;
+                    interval  = min_interval;
+                } else {
+                    pkt_burst = PKT_BURST_SCALE * min_interval / interval;
+                    interval  = min_interval;
+                }
+
+                bool     finish             = false;
+                int64_t  current_interval   = interval,
+                         interval_error     = 0;
+                uint64_t min_rtt            = UINT64_MAX,
+                         max_rtt            = 0,
+                         average_rtt        = 0,
+                         transmitted_count  = 0,
+                         received_count     = 0,
+                         transmitted_volume = 0,
+                         received_volume    = 0,
+                         total_count        = volume % pkt_size == 0 ? volume / pkt_size :
+                                                                       volume / pkt_size + 1,
+                         pkt_burst_error    = 0;
+
+                struct timespec start, end, report;
+
+                get_time(&start);
                 get_time(&end);
                 get_time(&report);
 
-                int64_t  current_interval = interval,
-                         interval_error   = 0;
-                uint64_t total_count      = volume % pkt_size == 0 ? volume / pkt_size :
-                                                                     volume / pkt_size + 1,
-                         pkt_burst_error  = 0;
-
                 while (!finish) {
-                    struct timespec start;
+                    struct timespec interval_start;
 
-                    get_time(&start);
+                    get_time(&interval_start);
 
-                    uint64_t pkt_count = total_count - transmitted_count > pkt_burst / PKT_BURST_PRECISION + pkt_burst_error / PKT_BURST_PRECISION ?
-                                                                           pkt_burst / PKT_BURST_PRECISION + pkt_burst_error / PKT_BURST_PRECISION :
+                    uint64_t pkt_count = total_count - transmitted_count > pkt_burst / PKT_BURST_SCALE + pkt_burst_error / PKT_BURST_SCALE ?
+                                                                           pkt_burst / PKT_BURST_SCALE + pkt_burst_error / PKT_BURST_SCALE :
                                                                            total_count - transmitted_count;
 
 #if defined(ENABLE_MMSG) && defined(HAVE_SENDMMSG)
                     sendmmsg_ping(ipv4_mode, sock, to_ai, pkt_size, ident, pkt_count, &transmitted_count, &transmitted_volume);
 #else
                     for (uint64_t i = 0; i < pkt_count; i++) {
-                        send_ping(ipv4_mode, sock, to_ai, pkt_size, ident, !i, &transmitted_count, &transmitted_volume);
+                        send_ping(ipv4_mode, sock, to_ai, pkt_size, ident, i == 0, &transmitted_count, &transmitted_volume);
                     }
 #endif
 
-                    pkt_burst_error  = pkt_burst_error % PKT_BURST_PRECISION;
-                    pkt_burst_error += pkt_burst       % PKT_BURST_PRECISION;
+                    pkt_burst_error  = pkt_burst_error % PKT_BURST_SCALE;
+                    pkt_burst_error += pkt_burst       % PKT_BURST_SCALE;
 
                     int64_t select_timeout = current_interval;
 
@@ -706,7 +782,9 @@ int main(int argc, char *argv[])
 
                         int n = select(sock + 1, &fds, NULL, NULL, &timeout);
 
-                        if (n > 0) {
+                        if (n < 0) {
+                            fprintf(stderr, "%s: select() failed: %s\n", prog_name, strerror(errno));
+                        } else if (n > 0) {
 #if defined(ENABLE_MMSG) && defined(HAVE_RECVMMSG)
                             while (recvmmsg_ping(ipv4_mode, sock, ident, &received_count, &received_volume, &min_rtt, &max_rtt, &average_rtt)) {
 #else
@@ -722,11 +800,11 @@ int main(int argc, char *argv[])
 
                         get_time(&now);
 
-                        if (ts_sub(&now, &start) >= current_interval) {
+                        if (ts_sub(&now, &interval_start) >= current_interval) {
                             if (transmitted_volume >= volume) {
                                 finish = true;
                             } else {
-                                interval_error += ts_sub(&now, &start) - current_interval;
+                                interval_error += ts_sub(&now, &interval_start) - current_interval;
 
                                 if (interval_error >= interval / 2) {
                                     current_interval  = interval / 2;
@@ -738,30 +816,32 @@ int main(int argc, char *argv[])
 
                             break;
                         } else {
-                            select_timeout = current_interval - ts_sub(&now, &start);
+                            select_timeout = current_interval - ts_sub(&now, &interval_start);
                         }
                     }
 
                     get_time(&end);
 
-                    if (reporting_period != 0 && end.tv_sec - report.tv_sec >= reporting_period) {
-                        printf("Periodic: pkts sent/rcvd: %" PRIu64 "/%" PRIu64 ", volume sent/rcvd: %" PRIu64 "/%" PRIu64 " bytes, time: %ld sec, speed: %" PRIu64 " kbps, rtt min/max/average: %" PRIu64 "/%" PRIu64 "/%" PRIu64 " ms\n",
-                               transmitted_count, received_count, transmitted_volume, received_volume, (long int)(end.tv_sec - begin.tv_sec),
-                               end.tv_sec - begin.tv_sec ? ((received_volume / (end.tv_sec - begin.tv_sec)) * 8) / 1000 : (received_volume * 8) / 1000,
+                    if (reporting_period > 0 && end.tv_sec - report.tv_sec >= reporting_period) {
+                        printf("Periodic: pkts sent/rcvd: %" PRIu64 "/%" PRIu64 ", volume sent/rcvd: %" PRIu64 "/%" PRIu64 " bytes, time: %ld sec,"
+                               " speed: %" PRIu64 " kbps, rtt min/max/average: %" PRIu64 "/%" PRIu64 "/%" PRIu64 " ms\n",
+                               transmitted_count, received_count, transmitted_volume, received_volume, (long int)(end.tv_sec - start.tv_sec),
+                               end.tv_sec - start.tv_sec > 0 ? received_volume / (end.tv_sec - start.tv_sec) * 8 / 1000 : received_volume * 8 / 1000,
                                min_rtt == UINT64_MAX ? 0 : min_rtt, max_rtt, average_rtt);
 
                         get_time(&report);
                     }
                 }
 
-                printf("Total: pkts sent/rcvd: %" PRIu64 "/%" PRIu64 ", volume sent/rcvd: %" PRIu64 "/%" PRIu64 " bytes, time: %ld sec, speed: %" PRIu64 " kbps, rtt min/max/average: %" PRIu64 "/%" PRIu64 "/%" PRIu64 " ms\n",
-                       transmitted_count, received_count, transmitted_volume, received_volume, (long int)(end.tv_sec - begin.tv_sec),
-                       end.tv_sec - begin.tv_sec ? ((received_volume / (end.tv_sec - begin.tv_sec)) * 8) / 1000 : (received_volume * 8) / 1000,
+                printf("Total: pkts sent/rcvd: %" PRIu64 "/%" PRIu64 ", volume sent/rcvd: %" PRIu64 "/%" PRIu64 " bytes, time: %ld sec,"
+                       " speed: %" PRIu64 " kbps, rtt min/max/average: %" PRIu64 "/%" PRIu64 "/%" PRIu64 " ms\n",
+                       transmitted_count, received_count, transmitted_volume, received_volume, (long int)(end.tv_sec - start.tv_sec),
+                       end.tv_sec - start.tv_sec > 0 ? received_volume / (end.tv_sec - start.tv_sec) * 8 / 1000 : received_volume * 8 / 1000,
                        min_rtt == UINT64_MAX ? 0 : min_rtt, max_rtt, average_rtt);
 
                 freeaddrinfo(to_ai);
             } else {
-                exit_val = EX_SOFTWARE;
+                exit_val = EX_NOHOST;
             }
         }
     }
